@@ -30,7 +30,9 @@ class Node:
         dependents,
         level,
         process_definitions_directory="javascript_processes",
+        user_defined_processes={},
     ):
+        self.user_defined_processes = user_defined_processes
         self.set_appropriate_class(process_id)
         self.variable_wrapper_string = uuid.uuid4().hex
         self.node_id = node_id
@@ -63,20 +65,27 @@ class Node:
         return f"Node {self.node_id} ({self.process_id})\n"
 
     def set_appropriate_class(self, process_id):
-        class_types_for_process = {
-            "reduce_dimension": ReduceDimensionNode,
-            "add_dimension": AddDimensionNode,
-            "load_collection": LoadCollectionNode,
-            "save_result": SaveResultNode,
-            "merge_cubes": MergeCubesNode,
-            "if": IfNode,
-            "apply": ApplyNode,
-            "count": CountNode,
-            "array_apply": ArrayApplyNode,
-            "array_filter": ArrayFilterNode,
-        }
-        if class_types_for_process.get(process_id):
-            self.__class__ = class_types_for_process[process_id]
+        if process_id in self.user_defined_processes:
+            self.__class__ = UserDefinedProcessNode
+        else:
+            class_types_for_process = {
+                "reduce_dimension": ReduceDimensionNode,
+                "add_dimension": AddDimensionNode,
+                "load_collection": LoadCollectionNode,
+                "save_result": SaveResultNode,
+                "merge_cubes": MergeCubesNode,
+                "if": IfNode,
+                "apply": ApplyNode,
+                "count": CountNode,
+                "array_apply": ArrayApplyNode,
+                "array_filter": ArrayFilterNode,
+                "aggregate_temporal_period": AggregateTemporalPeriodNode,
+                "apply_dimension": ApplyDimensionNode,
+                "aggregate_temporal": AggregateTemporalNode,
+                "resample_cube_temporal": ResampleCubeTemporalNode,
+            }
+            if class_types_for_process.get(process_id):
+                self.__class__ = class_types_for_process[process_id]
 
     def is_process_defined(self, process_id):
         try:
@@ -125,12 +134,6 @@ class Node:
 
     def write_process(self):
         process_definition = self.load_process_code()
-
-        #         if process_definition is None:
-        #             return f"""function {self.get_process_function_name()}(arguments) {{
-        #     const {{lmao}} = arguments;
-        #     return lmao + 42;
-        # }}"""
         return self.indent_by_level(process_definition)
 
     def write_call(self):
@@ -169,10 +172,9 @@ function reduce_dimension(arguments) {{
         return {self.child_nodes[-1].node_varname_prefix + self.child_nodes[-1].node_id};
     }}
 
-    const {{data, dimension}} = arguments; 
-    const newData = data.clone()
-    newData.reduceByDimension(reducer, dimension)
-    return newData;
+    {self.load_process_code()}
+    arguments['reducer'] = reducer;
+    return reduce_dimension(arguments);  
 }}
 """
 
@@ -199,6 +201,29 @@ class AddDimensionNode(Node):
 
 
 class MergeCubesNode(Node):
+    def write_process(self):
+        newline = "\n"
+        tab = "\t"
+        overlap_resolver_code = (
+            f"""
+function overlap_resolver(arguments) {{
+    {newline.join(node.write_function() for node in self.child_nodes)}
+    {newline.join(node.write_call() for node in self.child_nodes)}
+        return {self.child_nodes[-1].node_varname_prefix + self.child_nodes[-1].node_id};
+}}
+"""
+            if len(self.child_nodes) > 0
+            else "const overlap_resolver=undefined;"
+        )
+        return f"""
+function merge_cubes(arguments) {{
+    {overlap_resolver_code}
+    {self.load_process_code()}
+    arguments['overlap_resolver'] = overlap_resolver;
+    return merge_cubes(arguments);
+}}
+"""
+
     def get_dimensions_change(self, dimensions_of_inputs):
         """
         merge_cubes process expects datacubes to have different labels in a single dimension.
@@ -228,12 +253,8 @@ class IfNode(Node):
 
 
 class ApplyNode(Node):
-    def is_process_defined(self, process_id):
-        return True
-
     def write_process(self):
         newline = "\n"
-        tab = "\t"
         return f"""
 function apply(arguments) {{  
 
@@ -242,10 +263,29 @@ function apply(arguments) {{
     {newline.join(node.write_call() for node in self.child_nodes)}
         return {self.child_nodes[-1].node_varname_prefix + self.child_nodes[-1].node_id};
     }}
-    const {{data, dimension}} = arguments; 
-    const newData = data.clone()
-    newData.apply(process)
-    return newData;
+
+    {self.load_process_code()}
+    arguments['process'] = process;
+    return apply(arguments);
+
+}}
+"""
+
+
+class ApplyDimensionNode(Node):
+    def write_process(self):
+        newline = "\n"
+        return f"""
+function apply_dimension(arguments) {{
+   function process(arguments) {{
+    {newline.join(node.write_function() for node in self.child_nodes)}
+    {newline.join(node.write_call() for node in self.child_nodes)}
+        return {self.child_nodes[-1].node_varname_prefix + self.child_nodes[-1].node_id};
+    }}
+
+    {self.load_process_code()}
+    arguments['process'] = process;
+    return apply_dimension(arguments)
 }}
 """
 
@@ -269,8 +309,9 @@ function count(arguments) {{
 }}
 """
 
+
 class ArrayApplyNode(Node):
-    def write_process(self):   
+    def write_process(self):
         newline = "\n"
         tab = "\t"
         return f"""
@@ -287,6 +328,7 @@ function array_apply(arguments) {{
 }}
 """
 
+
 class ArrayFilterNode(Node):
     def write_process(self):
         newline = "\n"
@@ -303,3 +345,70 @@ function array_filter(arguments) {{
     return array_filter(arguments);
 }}
 """
+
+
+class AggregateTemporalPeriodNode(Node):
+    def write_process(self):
+        newline = "\n"
+        tab = "\t"
+        return f"""
+function aggregate_temporal_period(arguments) {{
+    function reducer(arguments) {{
+    {newline.join(node.write_function() for node in self.child_nodes)}
+    {newline.join(node.write_call() for node in self.child_nodes)}
+        return {self.child_nodes[-1].node_varname_prefix + self.child_nodes[-1].node_id};
+    }}
+
+    {self.load_process_code()}
+    arguments['reducer'] = reducer;
+    return aggregate_temporal_period(arguments)
+}}
+"""
+
+
+class AggregateTemporalNode(Node):
+    def write_process(self):
+        newline = "\n"
+        tab = "\t"
+        return f"""
+function aggregate_temporal(arguments) {{
+    function reducer(arguments) {{
+    {newline.join(node.write_function() for node in self.child_nodes)}
+    {newline.join(node.write_call() for node in self.child_nodes)}
+        return {self.child_nodes[-1].node_varname_prefix + self.child_nodes[-1].node_id};
+    }}
+
+    {self.load_process_code()}
+    arguments['reducer'] = reducer;
+    return aggregate_temporal(arguments);
+}}
+"""
+
+
+class ResampleCubeTemporalNode(Node):
+    def write_process(self):
+        newline = "\n"
+        tab = "\t"
+        return f"""
+function resample_cube_temporal(arguments) {{
+    {pkgutil.get_data("pg_to_evalscript", f"{self.process_definitions_directory}/is_valid.js").decode("utf-8")}
+
+    {self.load_process_code()}
+    return resample_cube_temporal(arguments);
+    }}
+"""
+
+
+class UserDefinedProcessNode(Node):
+    def is_process_defined(self, process_id):
+        return True
+
+    def write_process(self):
+        newline = "\n"
+        return f"""
+function {self.process_id}(arguments) {{  
+{newline.join(node.write_function() for node in self.child_nodes)}
+{newline.join(node.write_call() for node in self.child_nodes)}
+    return {self.child_nodes[-1].node_varname_prefix + self.child_nodes[-1].node_id};
+}}
+"""
